# -*- coding: utf-8 -*-
"""
lories.application.view.pages.footer
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from __future__ import annotations


# TODO Implement footer with "Powered by ISC Konstanz e.V." or similar
class PageFooter:
    pass
