/* set weather specific styles */
/*.weather-forecast {
  padding-top: 1rem;
  padding-bottom: 1rem;
  background-image: linear-gradient(#91d0ff, #c2e4ff)
  background-repeat: no-repeat;
  background-size: 100% calc(100% - 32px - 1rem);
}
*/
