# -*- coding: utf-8 -*-
"""
lories.application.view.pages.layout
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from .card import (  # noqa: F401
    PageCard,
    PageCardItem,
)
from .page import PageLayout  # noqa: F401
