# -*- coding: utf-8 -*-
"""
lories.application.view.pages.components
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""

from .page import ComponentPage  # noqa: F401
from .group import ComponentGroup  # noqa: F401
